
import org.lwjgl.PointerBuffer;
import org.lwjgl.Sys;
import org.lwjgl.opengl.GLContext;
import org.lwjgl.glfw.GLFWimage;

import org.lwjgl.glfw.GLFWMonitorCallback;
import org.lwjgl.glfw.GLFWWindowSizeCallback;
import org.lwjgl.glfw.GLFWKeyCallback;
import org.lwjgl.glfw.GLFWErrorCallback;
import org.lwjgl.glfw.Callbacks;

import org.lwjgl.LWJGLUtil;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.FloatBuffer;
import java.nio.ByteOrder;
import java.io.FileInputStream;

import java.nio.IntBuffer;

import static org.lwjgl.opengl.GL11.*;
import static org.lwjgl.opengl.GL13.*;
import static org.lwjgl.opengl.GL15.*;
import static org.lwjgl.opengl.GL20.*;
import static org.lwjgl.opengl.GL21.*;
import static org.lwjgl.opengl.GL30.*;

import static org.lwjgl.system.MemoryUtil.*;
import static org.lwjgl.glfw.GLFW.*;

//import static Jimgui.*;

public final class Test {

    // the "printable" key states.
    boolean[] keys = new boolean[256];
    boolean autoRot = true;

    private float width,height,aspect;
    private Mat4 perspective = new Mat4();
    private Mat4 view = new Mat4();
    private Mat4 vp = new Mat4();

	private Test() { }

	public static void main(String[] args) {

        Test test = new Test();

		try {
			test.run();
		} finally {
			glfwTerminate();
		}

	}

	private void run() {
				
		glfwSetErrorCallback(Callbacks.errorCallbackPrint(System.out));
		
		if ( glfwInit() == 0 )
			throw new IllegalStateException("Failed to initialize GLFW.");

        /*
         * create a window but request it be invisible for now
         */
        glfwDefaultWindowHints();
		glfwWindowHint(GLFW_VISIBLE, GL_FALSE);

	
        //glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);
        glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE); // requires OpenGL 3.2 or later

        glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3); // Use OpenGL Core v3.2
        glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 2);

		long window = glfwCreateWindow(640, 480, "Lwjgl test - No Gimbal!", NULL, NULL);
		if ( window == 0L )
			throw new IllegalStateException("Failed to create GLFW window.");

        /*
         * set the windows callback to be handled by this class and create
         * an openGL context for it
         */
		glfwSetWindowSizeCallback(window, new GLFWWindowSizeCallback() {
			@Override
			public void invoke(long window, int width, int height) {
				//printEvent("resized to %d x %d", window, width, height);
				updateProjection(width,height);
			}
		});
		
		glfwSetKeyCallback(window, new GLFWKeyCallback() {
			public void invoke(long window, int key, int scancode, int action, int mods) {
				switch ( action ) {
					case GLFW_RELEASE:
						if (key<255) keys[key]=false;
						break;
					case GLFW_PRESS:
						if (key<255) keys[key]=true;
						break;
				}

			}
		});

        glfwMakeContextCurrent(window);
		GLContext.createFromCurrent();

		
        // Once we have a view matrix make the projection matrix and
        // hence the view projection matrix
        Vec3 eye    =   new Vec3(0,0,0);
        Vec3 centre =   new Vec3(0,0,-5f);
        Vec3 up     =   new Vec3(0,1f,0);

        view.lookAt(eye, centre, up);
        updateProjection(640,480);

        // mouse cursor
        if ( LWJGLUtil.getPlatform() != LWJGLUtil.Platform.MACOSX ) {
			Util.loadMouseCursor(window, "data/cursor.png");
		}
        /*
         * load an image and create an openGL texture from it
         */
        int texture=0; 

        glActiveTexture(GL_TEXTURE0);

        texture = Util.loadTexture("data/wall.png",false);

        /*
         * vertex and fragment shaders
         * (you could load from file if you have a bunch of them!)
         */

        String fragShaderText =
//(for GLES)			"precision mediump float;\n"+
            "uniform sampler2D u_texture;\n"+
            "varying vec4 v_colour;\n"+
            "varying vec2 v_frag_uv; \n"+
            "void main() {\n"+
            "	vec4 baseColour = texture2D(u_texture,v_frag_uv);\n"+
            "   gl_FragColor = v_colour * baseColour;\n"+
//            "   gl_FragColor = vec4(1f,0f,0f,1f);\n"+
            "}\n";

        String vertShaderText =
            "uniform mat4 modelviewProjection;\n"+
            "attribute vec4 pos;\n"+
            "attribute vec2 uv_attrib;\n"+
            "attribute vec4 colour;\n"+
            "varying vec4 v_colour;\n"+
            "varying vec2 v_frag_uv;\n"+
            "void main() {\n"+
            "   gl_Position = modelviewProjection * pos ;\n"+
//            "   gl_Position = pos;\n"+
            "   v_colour = colour;\n"+
            "	v_frag_uv = uv_attrib;\n"+
            "}\n";

        
        int program = Util.createShaderProgram(vertShaderText,fragShaderText);
        glUseProgram(program);

        /*
         * 
         * get access to the shaders attibutes and uniforms
         *   
         */            
        int attr_pos=1,attr_colour=2,attr_uv=3,u_matrix,u_texture;
        ByteBuffer val = ByteBuffer.allocateDirect(4);
        
        glBindAttribLocation(program, attr_pos,     "pos");
        glBindAttribLocation(program, attr_colour,  "colour");
        glBindAttribLocation(program, attr_uv,      "uv_attrib");

        glLinkProgram(program);  /* needed to put attribs into effect */
        glGetProgrami(program, GL_LINK_STATUS, val);
        
        if (val.asIntBuffer().get(0)==0) {
            System.out.println("Shader log:\n"+glGetProgramInfoLog(program));
            System.exit(-1);
        }

        u_matrix = glGetUniformLocation(program, "modelviewProjection");
        u_texture = glGetUniformLocation(program, "u_texture");


        /*
         * 
         * Normally you'd load your "model" from a binary file...
         * here they are just in arrays
         */    

        /*
         * load the array into nio buffer
         */
        FloatBuffer vertsBuffer;
        vertsBuffer = ByteBuffer.allocateDirect(Shapes.pyramidVerts.length * 4).order(ByteOrder.nativeOrder())
                                    .asFloatBuffer().put(Shapes.pyramidVerts);
        vertsBuffer.flip();

        // make GL vertex array object and buffer from the nio floatbuffer
        int vbo2 = glGenBuffers();
        int vao2 = glGenVertexArrays();

        glBindVertexArray(vao2);
        glBindBuffer(GL_ARRAY_BUFFER, vbo2);
        glBufferData(GL_ARRAY_BUFFER, vertsBuffer , GL_STATIC_DRAW);



        vertsBuffer = ByteBuffer.allocateDirect(Shapes.boxVerts.length * 4).order(ByteOrder.nativeOrder())
                                    .asFloatBuffer().put(Shapes.boxVerts);
        vertsBuffer.flip();



        int vbo = glGenBuffers();
        int vao = glGenVertexArrays();

        glBindVertexArray(vao);
        glBindBuffer(GL_ARRAY_BUFFER, vbo);
        glBufferData(GL_ARRAY_BUFFER, vertsBuffer , GL_STATIC_DRAW);

        Mat4 mvp = new Mat4();		// Model View Projection
        Mat4 model = new Mat4();	// the objects rotation and position
        Mat4 mvpSave = new Mat4();	// for restoring mvp matrix (when rendering different bits with offsets)
        Mat4 tmpMat = new Mat4(); 	// temporary intermediate values

        /*
         * show the window and set the swap interval
         * as well as the background colour
         */
		glfwShowWindow(window);
		glfwSwapInterval(1);
		glClearColor(0.25f, 1f, 0.5f, 0.0f);
                
        /*
         * xa,ya,za is used when rotating with keyboard
         * x,y,z is the models position
         */
        float xa=0,ya=0,za=0;
        float x=0,y=0,z=-6;
        double thisTime = glfwGetTime();
        double lastTime = 0;
        float delta = 0;

        Quat tmpRot = new Quat();
        Quat modelRot = new Quat();

        final float ANG_SPEED = 1.75f;
        final float MOVE_SPEED = .5f;

        glEnable(GL_DEPTH_TEST);
        
        /*
         * The "main loop" for our demo
         */ 
		while ( glfwWindowShouldClose(window)==0 ) {

			glfwPollEvents();
                    

			xa=0;ya=0;za=0;

			// angle is an increment value
			// translate is a cumulative value
			if (keys[GLFW_KEY_Q]) xa =- ANG_SPEED * delta;
			if (keys[GLFW_KEY_W]) xa =+ ANG_SPEED * delta;
			if (keys[GLFW_KEY_A]) ya =- ANG_SPEED * delta;
			if (keys[GLFW_KEY_S]) ya =+ ANG_SPEED * delta;
			if (keys[GLFW_KEY_Z]) za =- ANG_SPEED * delta;
			if (keys[GLFW_KEY_X]) za =+ ANG_SPEED * delta;

			if (keys[GLFW_KEY_R]) x -= MOVE_SPEED * delta;
			if (keys[GLFW_KEY_T]) x += MOVE_SPEED * delta;
			if (keys[GLFW_KEY_F]) y -= MOVE_SPEED * delta;
			if (keys[GLFW_KEY_G]) y += MOVE_SPEED * delta;
			if (keys[GLFW_KEY_V]) z -= MOVE_SPEED * delta;
			if (keys[GLFW_KEY_B]) z += MOVE_SPEED * delta;
                

            glViewport(0, 0, (int)width, (int)height);
      
            /* the models position */
            model.translation( x, y, z );

            /* we take an angle increment and multiply it to our cumulative quaternion
             * as we are using an orientation (the quaternion) rather than 3 euler
             * rotation angles we avoid gimbal lock */
			
			// NB bacause xa,ya,za are delta values we can't just use 3 fixed
			// value quaternions here...  
			if (xa!=0) { tmpRot.axisAngle(1,0,0,xa); modelRot.multiply(tmpRot); }
			if (ya!=0) { tmpRot.axisAngle(0,1,0,ya); modelRot.multiply(tmpRot); }
			if (za!=0) { tmpRot.axisAngle(0,0,1,za); modelRot.multiply(tmpRot); }

            tmpMat.set(modelRot);
            model.multiply(tmpMat);
			
            /* the view/projection matrix is multiplied with the model matix
             * to give us the model, view, projection matrix for the shader */
            mvp.set(vp);
            mvp.multiply(model);

            glUniformMatrix4(u_matrix, false , mvp.mat);
       
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

            glEnableVertexAttribArray(attr_pos);
            glEnableVertexAttribArray(attr_colour);
            glEnableVertexAttribArray(attr_uv);

            // attrib, number items, type, normalised, byte stride, byte offset
            glVertexAttribPointer( attr_pos, 3, GL_FLOAT, false, 32, 0 );
            glVertexAttribPointer( attr_colour, 3, GL_FLOAT, false, 32, 12 );
            glVertexAttribPointer( attr_uv, 2, GL_FLOAT, false, 32, 24 );


            glBindVertexArray(vao);
            glBindBuffer(GL_ARRAY_BUFFER, vbo);
            glDrawArrays(GL_TRIANGLES, 0, 36);


            glBindVertexArray(vao2);
            glBindBuffer(GL_ARRAY_BUFFER, vbo2);


            final int pyrVerts = 6*3;
            mvpSave.set(mvp); // save current state
            
            tmpMat.translation(0,0,1.2f); // (set to a translation NOT translated)
            mvp.multiply(tmpMat);
            tmpMat.rotationX(1.5707963268f); // 90 degrees
            mvp.multiply(tmpMat);
            glUniformMatrix4(u_matrix, false , mvp.mat);
            glDrawArrays(GL_TRIANGLES, 0, pyrVerts);

            tmpMat.translation(1.2f,0,0);
            mvp.set(mvpSave);
            mvp.multiply(tmpMat);
            tmpMat.rotationZ(-1.5707963268f); // 90 degrees
            mvp.multiply(tmpMat);
            glUniformMatrix4(u_matrix, false , mvp.mat);
            glDrawArrays(GL_TRIANGLES, 0, pyrVerts);

            tmpMat.translation(0,1.2f,0);
            mvp.set(mvpSave);
            mvp.multiply(tmpMat);
            glUniformMatrix4(u_matrix, false , mvp.mat);
            glDrawArrays(GL_TRIANGLES, 0, pyrVerts);


            glDisableVertexAttribArray(attr_pos);
            glDisableVertexAttribArray(attr_colour);
            glDisableVertexAttribArray(attr_uv);
          
			glfwSwapBuffers(window);

            lastTime = thisTime;
            thisTime = glfwGetTime();
            delta = (float)(thisTime-lastTime);
		}


	}


    private void updateProjection( float w, float h) {
        width = w; height = h;
        aspect = w / h;

        perspective.perspectiveProjection(45f, aspect, .1f, 10f);

        // as a view is stationary we can make the View * Perspective
        // matrix for use with the model matrix
        vp.set(perspective);
        vp.multiply(view);
    }

}
