import java.nio.ByteBuffer;
import java.io.IOException;
import java.io.FileInputStream;

//import org.lwjgl.system.glfw.GLFWimage;
import org.lwjgl.glfw.GLFWimage;
//import static org.lwjgl.system.glfw.GLFW.*;
import static org.lwjgl.glfw.GLFW.*;
import static org.lwjgl.opengl.GL11.*;          // core GL
import static org.lwjgl.opengl.GL20.*;          // GL_FRAGMENT_SHADER etc

public class Util {

    Util() {    }

    public static void loadMouseCursor(long window, String fileName) {         
        /*
         * load and set a custom cursor just for our window
         */
        try {
            FileInputStream in = new FileInputStream(fileName);
            PNGDecoder decoder = new PNGDecoder(in);
            // assuming RGBA here but should allow for RGB and RGBA
            ByteBuffer buf = ByteBuffer.allocateDirect(4*decoder.getWidth()*decoder.getHeight());
            decoder.decode(buf, decoder.getWidth()*4, PNGDecoder.Format.RGBA);
            buf.flip();
            
            ByteBuffer img = GLFWimage.malloc(decoder.getWidth(), decoder.getHeight(), buf);
            long cursor = glfwCreateCursor(img, 0, 8);
            glfwSetCursor(window, cursor);
		} catch (IOException e) {
			e.printStackTrace();
		}
    }


    public static int loadTexture(String fileName, boolean withAlpha) {
        int texture=0;
        try {
            FileInputStream in = new FileInputStream(fileName);
            PNGDecoder decoder = new PNGDecoder(in);

            int sz=3;
            int gf=GL_RGB;
            PNGDecoder.Format pf=PNGDecoder.Format.RGB;
            if (withAlpha) {
                sz=4;
                gf=GL_RGBA;
                pf=PNGDecoder.Format.RGBA;
            }
            ByteBuffer buf = ByteBuffer.allocateDirect(sz*decoder.getWidth()*decoder.getHeight());
            decoder.decode(buf, decoder.getWidth()*sz, pf);
            buf.flip();

            texture=glGenTextures();
            glBindTexture(GL_TEXTURE_2D, texture);
            glTexImage2D(GL_TEXTURE_2D, 0, gf, decoder.getWidth(), decoder.getHeight(), 0,
                    gf, GL_UNSIGNED_BYTE, buf);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
            in.close();
        } catch (IOException e) {
            e.printStackTrace();   
        }
        return texture;
    }



    public static int createShaderProgram(String vertShaderText, String fragShaderText) {
        int fragShader, vertShader, program=0;
        int stat;
         // val is a byte buffer which is reused for various (single) int return values
        final ByteBuffer val = ByteBuffer.allocateDirect(4); 
               
        fragShader = glCreateShader(GL_FRAGMENT_SHADER);
        glShaderSource(fragShader, fragShaderText);
        glCompileShader(fragShader);
        glGetShaderi(fragShader, GL_COMPILE_STATUS, val);
        if (val.asIntBuffer().get(0)==0) {
            System.out.println("Error: fragment shader did not compile!\n");
            System.out.println("Shader log:\n"+glGetShaderInfoLog(fragShader));
            System.exit(-1);
        }
        
        vertShader = glCreateShader(GL_VERTEX_SHADER);
        glShaderSource(vertShader, vertShaderText);
        glCompileShader(vertShader);
        glGetShaderi(vertShader, GL_COMPILE_STATUS, val);
        if (val.asIntBuffer().get(0)==0) {
            System.out.println("Error: vertex shader did not compile!\n");
            System.out.println("Shader log:\n"+glGetShaderInfoLog(vertShader));
            System.exit(-1);
        }       

        /*
         * link the shaders into a shader "program"
         */
        program = glCreateProgram();
        glAttachShader(program, fragShader);
        glAttachShader(program, vertShader);
        glLinkProgram(program);
        
        glGetProgrami(program, GL_LINK_STATUS, val);
        if (val.asIntBuffer().get(0)==0) {
            System.out.println("Shader log:\n"+glGetProgramInfoLog(program));
            System.exit(-1);
        }

        return program;
    }

    public static void checkError(String tag) {
        int err = glGetError();
        if (err!=GL_NO_ERROR) {
            System.out.print("tag "+tag+" error:");
            if (err==GL_INVALID_ENUM) System.out.println("Invalid enum\n");
            if (err==GL_INVALID_VALUE) System.out.println("invalid value\n");
            if (err==GL_INVALID_OPERATION) System.out.println("invalid operation\n");
            if (err==GL_STACK_OVERFLOW) System.out.println("stack overflow\n");
            if (err==GL_STACK_UNDERFLOW) System.out.println("stack underflow\n");
            if (err==GL_OUT_OF_MEMORY) System.out.println("out of memory\n");
            System.exit(-1);
        }
    }
}
